@@ -8,7 +8,7 @@ use IO::Socket qw(:crlf);
 use HTTP::Parser::XS qw(parse_http_request);
 use HTTP::Status qw(status_message);
 use HTTP::Date qw(time2str);
-use POSIX qw(EINTR);
+use POSIX qw(EINTR EPIPE);
 use Symbol;
 
 use Plack::Util;
@@ -540,16 +540,22 @@ sub _finalize_response {
 sub _syswrite {
     my ($conn, $buffer_ref) = @_;
 
-    while (length $$buffer_ref) {
-        my $len = syswrite($conn, $$buffer_ref);
+    my $amount = length $$buffer_ref;
+    my $offset = 0;
 
-        if (defined $len) {
-            substr($$buffer_ref, 0, $len, '');
-            DEBUG && warn "[$$] Wrote $len byte", ($len == 1 ? '' : 's'), "\n";
-        }
-        elsif ($! != EINTR) {
+    while ($amount > 0) {
+        my $len = syswrite($conn, $$buffer_ref, $amount, $offset);
+
+        if (not defined $len) {
+            return if $! == EPIPE;
+            redo if $! == EINTR;
             die "write error: $!";
         }
+
+        $amount -= $len;
+        $offset += $len;
+
+        DEBUG && warn "[$$] Wrote $len byte", ($len == 1 ? '' : 's'), "\n";
     }
 }
 
